package yapl4.interfaces;

import yapl4.lib.Type;

/**
 * Attributes of production symbols for type checking and generating
 * 3-address-code. An object implementing this interface stores attributes of production
 * symbols of the YAPL4 grammar. Essentially, an Attrib object represents one
 * operand of a 3-address-code instruction, or a branch condition.
 * <p>
 * The constants defined by this interface specify the various kinds of operands
 * represented by an object implementing this interface.
 * </p>
 * 
 * @author Mario Taschwer
 * @version $Id$
 */
public interface Attrib {
	/** Invalid operand. */
	public static final byte None = 0;
	
	/** Register operand. Requires a data type and a register number. */
	public static final byte Register  = 1;
	
	/** Constant operand. Requires a data type. */
	public static final byte Constant  = 2;
	
	/** Variable (word value) in memory. Requires a data type and a memory address. 
	 * The address points to a memory location storing the variable value
	 * of the given data type.
	 */
	public static final byte Variable  = 3;
	
	/** Address operand (used e.g. with arrays). 
	 * Requires a data type and a memory address.
	 * The address points to a memory location storing the target object
	 * of the given data type.
	 */
	public static final byte Address   = 4;
	
	/** Address register operand.
	 * Requires a register number and a data type.
	 * The register holds the (absolute) memory address of the target object
	 * of the given data type.
	 */
	public static final byte AddrRegister = 5;
	
	/** Memory reference. Requires a memory address and a data type. 
	 * The address points to a memory location storing the 
	 * (absolute) address of the target object of the given data type.
	 */
	public static final byte Reference = 6;

	/** Return the kind of operand represented by this object.
	 * 
	 * @return one of the constants defined by this interface.
	 */
	public byte getKind();
	
	/**
	 * Set the kind of operand represented by this object.
	 * 
	 * @param kind     one of the constants defined by this interface.
	 */
	public void setKind(byte kind);
	
	/** Get the data type of this operand (or its target object). */
	public Type getType();
	
	/** Set the data type of this operand (or its target object). */
	public void setType(Type type);
	
	/** Return <code>true</code> iff this operand is constant. */
	public boolean isConstant();
	
	/** Specify whether this operand is constant. */
	public void setConstant(boolean isConstant);
	
	/** Return <code>true</code> iff this operand represents
	 * a global memory object (i.e. it is stored in heap memory).
	 */
	public boolean isGlobal();
	
	/** Specify whether this operand represents a global memory object.
	 * @see #isGlobal()
	 */
	public void setGlobal(boolean isGlobal);
	
	/** Return the address offset (for memory operands).
	 *  If this operand represents a global memory object
	 *  (see {@link #isGlobal()}), the offset is relative to
	 *  the global variable storage area in heap memory. Otherwise,
	 *  the offset is relative to the current stack frame.
	 */
	public int getOffset();
	
	/** Set the address offset (for memory operands). 
	 * @see #getOffset()
	 */
	public void setOffset(int offset);
	
	/** Get the register number (for register operands). */
	public int getRegister();
	
	/** Set the register number (for register operands). */
	public void setRegister(int register);
}
